"""
CCXW - CryptoCurrency eXchange Websocket Library
Common functions used by the package

Author: Ricardo Marcelo Alvarez
Date: 2023-10-31
"""

import urllib.parse
import urllib.request
import json
import socket

def is_json(myjson):
    """
    is_json
    =======
        This function get a string or bytes and check if json return True
        if the input is a json valid string.
            :param myjson: str | bytes.

            :return bool: Return True if myjson is a str and is a json. 
    """

    result = False

    if myjson is not None and isinstance(myjson,(str,bytes)):
        try:
            json.loads(myjson)
            result = True
        except Exception: # pylint: disable=broad-except
            result = False

    return result

def is_convertible_to_json(myvar):
    """
    is_convertible_to_json
    ======================
        This function check if myvar is convertible to json string.
            :param myvar:

            :return bool: Return True if myvar is convertible to json string.
    """

    result = False
    try:
        json.dumps(myvar)
        result = True
    except Exception: # pylint: disable=broad-except
        result = False

    return result

def file_get_contents_url(url,mode="b",post_data=None,headers=None,timeout=9):
    """
    file_get_contents_url
    =====================
        This function get a url and reads into a string
            :param url: str file URL.
            :param mode: str "b" for binary response.
            :param post_data: dict Post data in format key -> value.
            :param headers: dict headers in format key -> value.
            :param timeout: int request timeout.

            :return str: Return response data from url. 
    """

    result = None

    if headers is None:
        headers = {}

    try:
        req = None
        if post_data is not None and isinstance(post_data,dict):
            req = urllib.request.Request(url, urllib.parse.urlencode(post_data).encode(),headers)
        else:
            req = urllib.request.Request(url, None,headers)

        if req is not None:
            try:
                with urllib.request.urlopen(req, None, timeout=timeout) as response:
                    result = response.read()

            except Exception: # pylint: disable=broad-except
                result = None

        if result is not None and isinstance(result,bytes):
            result = result.decode()

    except Exception: # pylint: disable=broad-except
        result = None

    if mode != "b" and result is not None and result is not False and isinstance(result,bytes):
        result = result.decode()

    return result

def is_port_free(port, host='localhost'):
    """
    is_port_free
    ============
        This function check if port already used
            :param port: int port number.
            :param host: str 

            :return bool: Return True if port is free
    """
    result = False

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((host, port))
        s.close()
        result = True
    except Exception: # pylint: disable=broad-except
        result = False

    return result



"""
Ccxw - CryptoCurrency eXchange Websocket Library
Kucoin auxiliary functions

Author: Ricardo Marcelo Alvarez
Date: 2023-10-31
"""

import os
import json
import time
import datetime
import queue
import math
import copy

import ccxw.ccxw_common_functions as ccf
from ccxw.safe_thread_vars import DictSafeThread
import ccxw

class KucoinCcxwAuxClass():
    """
    Ccxw - CryptoCurrency eXchange Websocket Library KucoinCcxwAuxClass
    ===================================================================
        This class contains helper functions for the Ccxw class.
    """

    def __init__(self,\
                 streams: list[dict],\
                 trading_type: str='SPOT',\
                 testmode: bool=False,\
                 result_max_len: int=5,\
                 data_max_len: int=1000,\
                 debug: bool=False):
        """
        KucoinCcxwAuxClass constructor
        ==============================
            :param self: KucoinCcxwAuxClass instance.
            :param streams: list[dict]
                                    dicts must have this struct.
                                        {
                                            'endpoint': str only allowed 'order_book' | 'kline' |\
                                                'trades' | 'ticker',
                                            'symbol': str unified symbol.,
                                            'interval': str '1m' | '3m' | '5m' | '15m' | '30m' |\
                                                '1h' | '2h' | '4h' | '6h' | '8h' | '12h' | '1d' |\
                                                '3d' | '1w' | '1mo' for 'kline' endpoint is\
                                                    mandatory.
                                        }

            :param trading_type: str only allowed 'SPOT'.
            :param testmode: bool.
            :param result_max_len: int Max return values > 1 and <= data_max_len.
            :param data_max_len: int. > 1 and <= 400 max len of data getting from exchange.
            :param debug: bool Verbose output.

            :return: Return a new instance of the Class KucoinCcxwAuxClass.
        """

        __exchange_limit_streams = 100

        self.__exchange = os.path.basename(__file__)[:-3]
        self.__ws_streams = streams
        self.__testmode = testmode

        self.__exchange_info_cache = {}
        self.__exchange_info_cache['data'] = None
        self.__exchange_info_cache['last_get_time'] = 0

        self.__debug = debug # pylint: disable=unused-private-member

        self.__trading_type = trading_type
        self.__data_max_len = data_max_len

        self.__data_max_len = min(self.__data_max_len, 400)
        self.__data_max_len = max(self.__data_max_len, 1)

        self.__result_max_len = result_max_len

        self.__result_max_len = min(self.__result_max_len, 500)
        self.__result_max_len = max(self.__result_max_len, 1)

        self.__ws_url_api = None
        self.__ws_url_test = None
        self.__ws_url = None # pylint: disable=unused-private-member
        self.__url_api = None
        self.__url_test = None
        self.__ws_endpoint_url = None
        self.__ws_endpoint_on_open_vars = None
        self.__ws_endpoint_on_close_vars = None
        self.__ws_temp_data = DictSafeThread()

        self.ping_interval_ms = 10.0
        self.ping_timeout_ms = 10.0

        if not self.__check_streams_struct(streams):
            raise ValueError('The streams struct is not valid' + str(streams))

        if len(streams) > __exchange_limit_streams:
            raise ValueError('The exchange ' + str(self.__exchange)\
                             + ' not alowed more than ' + str(__exchange_limit_streams)\
                             + ' of streams.')

    def __del__(self):
        pass

    def __check_streams_struct(self, streams):
        """
        __check_streams_struct
        ======================
            This function check streams struct
                :param self: This class instance.
                :param streams: list[dict]
                                dicts must have this struct.
                                    {
                                        'endpoint': str only allowed 'order_book' | 'kline' |\
                                            'trades' | 'ticker',
                                        'symbol': str unified symbol.,
                                        'interval': str '1m' | '3m' | '5m' | '15m' | '30m' |\
                                            '1h' | '2h' | '4h' | '6h' | '8h' | '12h' | '1d' |\
                                            '3d' | '1w' | '1mo' only for 'kline' endpoint is\
                                                mandatory.
                                    }

                :return bool: True if streams struct is correct otherwise False
        """
        result = False

        if streams is not None and isinstance(streams, list) and len(streams) > 0:
            result = True
            for stream in streams:
                result = result and stream is not None\
                    and isinstance(stream, dict)\
                    and 'endpoint' in stream\
                    and stream['endpoint'] is not None\
                    and isinstance(stream['endpoint'], str)\
                    and stream['endpoint'] in ccxw.Ccxw.get_supported_endpoints()\
                    and 'symbol' in stream\
                    and stream['symbol'] is not None\
                    and isinstance(stream['symbol'], str)\
                    and self.if_symbol_supported(stream['symbol'])

                if result and stream['endpoint'] == 'kline':
                    result = 'interval' in stream\
                        and stream['interval'] is not None\
                        and isinstance(stream['interval'], str)\
                        and stream['interval'] in\
                            ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h',\
                             '6h', '8h', '12h', '1d', '1w']

        return result

    def get_websocket_url(self):
        """
        get_websocket_url
        =================
            This function set and return websocket URL.

                :return str: Return websocket URL.
        """

        result = None

        self.get_api_url()
        __data_token_raw = None

        if self.__testmode:
            if self.__url_test is not None:
                __url_token = self.__url_test + '/api/v1/bullet-public'
                __data_token_raw = ccf.file_get_contents_url(__url_token,'b',{})
        else:
            if self.__url_api is not None:
                __url_token = self.__url_api + '/api/v1/bullet-public'
                __data_token_raw = ccf.file_get_contents_url(__url_token,'b',{})

        if __data_token_raw is not None and ccf.is_json(__data_token_raw):
            __data_token = json.loads(__data_token_raw)

            if isinstance(__data_token,dict) and 'data' in __data_token\
                and isinstance(__data_token['data'],dict):
                __comp_0 = 'token' in __data_token['data']\
                    and isinstance(__data_token['data']['token'],str)\
                    and len(__data_token['data']['token']) > 0
                __comp_1 = 'instanceServers' in __data_token['data']\
                    and isinstance(__data_token['data']['instanceServers'],list)\
                    and len(__data_token['data']['instanceServers']) > 0\
                    and 'endpoint' in __data_token['data']['instanceServers'][0]\
                    and isinstance(__data_token['data']['instanceServers'][0]['endpoint'],str)\
                    and len(__data_token['data']['instanceServers'][0]['endpoint']) > 0

                if __comp_0 and __comp_1:
                    self.__ws_url_api = __data_token['data']['instanceServers'][0]['endpoint']
                    __ws_token = __data_token['data']['token']
                    self.__ws_url_api = self.__ws_url_api + '?token=' + __ws_token
                    self.__ws_url_test = self.__ws_url_api

                    if 'pingInterval' in __data_token['data']['instanceServers'][0]:
                        self.ping_interval_ms = (
                            float(int(__data_token['data']['instanceServers'][0]['pingInterval'])\
                                  / 1000)
                        )

                    if 'pingTimeout' in __data_token['data']['instanceServers'][0]:
                        self.ping_timeout_ms = (
                            float(int(__data_token['data']['instanceServers'][0]['pingTimeout'])\
                                  / 1000)
                        )

            result = self.__ws_url_api

            if self.__testmode:
                result = self.__ws_url_test

        return result

    def get_api_url(self):
        """
        get_api_url
        ===========
            This function set and return API URL.

                :return str: Return API URL.
        """

        result = None

        if self.__trading_type == 'SPOT':
            self.__url_api = 'https://api.kucoin.com'
            self.__url_test = 'https://openapi-sandbox.kucoin.com'

        result = self.__url_api
        if self.__testmode:
            result = self.__url_test

        return result

    def get_exchange_info(self):
        """
        get_exchange_info
        =================
            This function get exchange info. 

                :return dict: Return exchange info.
        """

        result = None

        max_last_get_time = 7200

        current_time = int(time.time())

        if (current_time - self.__exchange_info_cache['last_get_time']) >= max_last_get_time:
            self.__exchange_info_cache['data'] = None

        if self.__exchange_info_cache['data'] is None:

            __l_url_api = self.get_api_url()

            __l_endpoint = '/api/v2/symbols'

            __l_url_point = __l_url_api + __l_endpoint

            __data = ccf.file_get_contents_url(__l_url_point)

            if __data is not None and ccf.is_json(__data):
                result = json.loads(__data)
                self.__exchange_info_cache['data'] = result
                self.__exchange_info_cache['last_get_time'] = current_time

        else:
            result = self.__exchange_info_cache['data']

        return result

    def get_exchange_full_list_symbols(self, sort_list=True):
        """
        get_exchange_full_list_symbols
        ==============================
            This function get exchange info. 
                :param sort_list: bool.
                :return dict: Return exchange info.
        """

        result = None
        __main_data = self.get_exchange_info()

        if __main_data is not None and isinstance(__main_data,dict)\
            and 'data' in __main_data and isinstance(__main_data['data'],list):

            result = []
            for symbol_data in __main_data['data']:
                if symbol_data is not None and isinstance(symbol_data,dict):
                    if 'baseCurrency' in symbol_data and 'quoteCurrency' in symbol_data\
                        and isinstance(symbol_data['baseCurrency'],str)\
                        and isinstance(symbol_data['quoteCurrency'],str):
                        result.append(\
                            symbol_data['baseCurrency'].upper() + '/'\
                            + symbol_data['quoteCurrency'].upper())

            if sort_list:
                result.sort()

        return result

    def get_unified_symbol_from_symbol(self, symbol):
        """
        get_unified_symbol_from_symbol
        ==============================
            This function get unified symbol from symbol.
                :param self: This class instance.
                :param symbol: str.
                :return str: Return unified symbol.
        """
        result = symbol

        full_list_symbols = self.get_exchange_full_list_symbols(False)

        for symbol_rpl in full_list_symbols:
            if symbol.replace('-', '').lower() == symbol_rpl.replace('/', '').lower():
                result = symbol_rpl
                break

        return result

    def __get_interval_from_unified_interval(self, interval):
        result = '1'

        if interval is not None:
            result = str(interval)

            if 'm' in result:
                result = result.replace('m', 'min')
            elif 'h' in result or 'H' in result:
                result = result.lower().replace('h', 'hour')
                result = 60 * int(result)
            elif 'd' in result or 'D' in result:
                result = result.lower().replace('d', 'day')
            elif 'w' in result or 'W' in result:
                result = result.lower().replace('w', 'week')

            result = str(result)

        return result

    def get_unified_interval_from_interval(self, interval):
        """
        get_unified_interval_from_interval
        ==================================
            This function get unified interval from interval.
                :param self: This class instance.
                :param interval: str.
                :return str: Return unified interval.
        """

        result = interval

        if result is not None:
            result = str(interval)

            if 'min' in result:
                result = result.replace('min', 'm')
            elif 'hour' in result:
                result = result.replace('hour', 'h')
            elif 'day' in result:
                result = result.replace('day', 'd')
            elif 'week' in result:
                result = result.replace('week', 'w')

            result = str(result)

        return result

    def if_symbol_supported(self, symbol):
        """
        if_symbol_supported
        ===================
            This function check if symbol is supported by the exchange.

                :return bool: Return True if supported 
        """

        result = False

        __data = self.get_exchange_full_list_symbols()

        if isinstance(__data,list) and symbol in __data:
            result = True

        return result

    def get_stream_index(self, endpoint, symbol, interval: str='none'):
        """
        get_stream_index
        ================
            This function return a stream index for use in dict.
                :param self: This class instance.
                :param endpoint: str.
                :param symbol: str unified symbol.
                :param interval: str.
                :return str: Return stream index
        """
        result = 'stream'

        if interval is None:
            interval = 'none'

        result = result + '_' + endpoint + '_'\
            + self.get_unified_symbol_from_symbol(symbol) + '_'\
            + self.get_unified_interval_from_interval(interval)

        result = result.replace('/','').lower()

        return result

    def get_websocket_endpoint_path(self):
        """
        get_websocket_endpoint_path
        ===========================
            This function set endpoint URL and on_open vars and on_close vars.

                :return dict: Return dict with endpoint, open_vars and close_vars
        """

        result = None

        self.__ws_endpoint_on_open_vars = []
        self.__ws_endpoint_on_close_vars = []

        self.__ws_endpoint_url = ''
        __id = str(time.time_ns())
        self.__ws_endpoint_url = '&connectId=' + __id

        for stream in self.__ws_streams:
            interval = 'none'

            __topic_open = None
            __topic_open = {}
            __topic_open['id'] = __id
            __topic_open['type'] = 'subscribe'
            __topic_open['topic'] = ''
            __topic_open['privateChannel'] = False
            __topic_open['response'] = True

            __topic_close = None
            __topic_close = copy.copy(__topic_open)
            __topic_close['type'] = 'unsubscribe'
            time.sleep(0.050)

            if stream['endpoint'] == 'order_book':
                result = True

                __topic_open['topic'] = '/market/level2:'\
                    + stream['symbol'].replace("/","-").upper()
                __topic_close['topic'] = __topic_open['topic']

            elif stream['endpoint'] == 'kline':
                interval = stream['interval']

                __topic_open['topic'] = '/market/candles:'\
                    + stream['symbol'].replace("/","-").upper()\
                    + '_' + self.__get_interval_from_unified_interval(interval)
                __topic_close['topic'] = __topic_open['topic']


            elif stream['endpoint'] == 'trades':
                __topic_open['topic'] = '/market/match:' + stream['symbol'].replace("/","-").upper()
                __topic_close['topic'] = __topic_open['topic']

            elif stream['endpoint'] == 'ticker':
                __topic_open['topic'] = '/market/ticker:'\
                    + stream['symbol'].replace("/","-").upper()
                __topic_close['topic'] = __topic_open['topic']

            __stream_index = self.get_stream_index(stream['endpoint'],\
                                                     stream['symbol'],\
                                                     interval=interval)
            self.__ws_temp_data[__stream_index] = None
            self.__ws_endpoint_on_open_vars.append(json.dumps(__topic_open))
            self.__ws_endpoint_on_close_vars.append(json.dumps(__topic_close))


        result = {}
        result['ws_endpoint_url'] = self.__ws_endpoint_url
        result['ws_endpoint_on_open_vars'] = self.__ws_endpoint_on_open_vars
        result['ws_endpoint_on_close_vars'] = self.__ws_endpoint_on_close_vars
        result['ws_ping_interval'] = self.ping_interval_ms
        result['ws_ping_timeout'] = self.ping_timeout_ms

        return result

    def __init_order_book_data(self, temp_data):

        result = False
        __data = None

        if 'symbol' in temp_data\
            and temp_data['symbol'] is not None\
            and isinstance(temp_data['symbol'], str):
            __symbol = temp_data['symbol']

            __url_dest = self.__url_api + '/api/v1/market/orderbook/level2_100?symbol='\
                + str(__symbol).replace("/","-").upper()


            __data = ccf.file_get_contents_url(__url_dest)

            if __data is not None and ccf.is_json(__data):
                __data = json.loads(__data)

                if __data is not None and isinstance(__data, dict)\
                    and 'code' in __data and 'data' in __data\
                    and __data['code'] == '200000':
                    __stream_index = self.get_stream_index('order_book', __symbol)
                    self.__ws_temp_data[__stream_index] = None
                    self.__ws_temp_data[__stream_index] = {}
                    self.__ws_temp_data[__stream_index]['endpoint'] = 'order_book'
                    self.__ws_temp_data[__stream_index]['exchange'] = self.__exchange
                    self.__ws_temp_data[__stream_index]['symbol'] = __symbol
                    self.__ws_temp_data[__stream_index]['interval'] = None
                    self.__ws_temp_data[__stream_index]['last_update_id'] = (
                        int(__data['data']['sequence'])
                    )
                    self.__ws_temp_data[__stream_index]['diff_update_id'] = 0
                    self.__ws_temp_data[__stream_index]['bids'] = __data['data']['bids']
                    self.__ws_temp_data[__stream_index]['asks'] = __data['data']['asks']
                    self.__ws_temp_data[__stream_index]['type'] = 'snapshot'

                    result = True

        return result

    def manage_websocket_message_order_book(self, data):
        """
        manage_websocket_message_order_book
        ===================================
            This function manage websocket message and normalize result
            data for order_book endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        if __temp_data is not None and isinstance(__temp_data,dict)\
            and 'type' in __temp_data and 'data' in __temp_data\
            and isinstance(__temp_data['data'],dict):
            if __temp_data['type'] == 'message':
                __symbol = __temp_data['topic'].split(':')[1]
                __stream_index = self.get_stream_index('order_book', __symbol)

                if 'changes' in __temp_data['data']\
                    and isinstance(__temp_data['data']['changes'],dict):
                    if self.__ws_temp_data[__stream_index] is None:
                        if self.__init_order_book_data(__temp_data['data'])\
                            and self.__manage_websocket_diff_data(__temp_data['data']):
                            __proc_data = True
                    elif self.__manage_websocket_diff_data(__temp_data['data']):
                        __proc_data = True

                if __proc_data:
                    __message_out = None
                    __message_out = {}
                    __message_out['endpoint'] = 'order_book'
                    __message_out['exchange'] = self.__exchange
                    __message_out['symbol'] = __symbol
                    __message_out['interval'] = None
                    __message_out['last_update_id'] = (
                        self.__ws_temp_data[__stream_index]['last_update_id']
                    )
                    __message_out['diff_update_id'] = (
                        self.__ws_temp_data[__stream_index]['diff_update_id']
                    )
                    __message_out['bids'] = (
                        self.__ws_temp_data[__stream_index]['bids'][:self.__result_max_len]
                    )
                    __message_out['asks'] = (
                        self.__ws_temp_data[__stream_index]['asks'][:self.__result_max_len]
                    )
                    __message_out['type'] = self.__ws_temp_data[__stream_index]['type']
                    __current_datetime = datetime.datetime.utcnow()
                    __current_timestamp = __current_datetime.strftime("%s.%f")
                    __current_datetime = (
                        __current_datetime.strftime("%Y-%m-%d %H:%M:%S.%f")
                    )
                    __message_out['timestamp'] = __current_timestamp
                    __message_out['datetime'] = __current_datetime

                    result = __message_out

        return result

    def manage_websocket_message_kline(self, data):
        """
        manage_websocket_message_kline
        ==============================
            This function manage websocket message and normalize result
            data for kline endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        __symbol = None
        __interval = None

        if __temp_data is not None and isinstance(__temp_data, dict):
            __interval = __temp_data['topic'].split('_')[1]

            if 'type' in __temp_data and 'data' in __temp_data\
                and isinstance(__temp_data['data'],dict):
                if 'candles' in __temp_data['data']\
                    and isinstance(__temp_data['data']['candles'],list)\
                    and len(__temp_data['data']['candles']) > 0\
                    and 'symbol' in __temp_data['data']:
                    __symbol = __temp_data['data']['symbol']
                    __stream_index = self.get_stream_index('kline', __symbol, __interval)

                    __delta_time = 60000

                    if 'min' in __interval:
                        __delta_mult = int(__interval.replace('min',''))
                        __delta_time = 60000 * __delta_mult

                    elif 'hour' in __interval:
                        __delta_mult = int(__interval.replace('hour',''))
                        __delta_time = 60000 * __delta_mult * 60

                    elif 'day' in __interval:
                        __delta_mult = int(__interval.replace('day',''))
                        __delta_time = 60000 * __delta_mult * 60 * 24

                    elif 'week' in __interval:
                        __delta_mult = int(__interval.replace('week',''))
                        __delta_time = 60000 * __delta_mult * 60 * 24

                    __delta_time = __delta_time - 1

                    if self.__ws_temp_data[__stream_index] is None\
                        or not isinstance(self.__ws_temp_data[__stream_index], dict):
                        self.__ws_temp_data[__stream_index] = {}

                    __message_add = None
                    __message_add = {}
                    __message_add['endpoint'] = 'kline'
                    __message_add['exchange'] = self.__exchange
                    __message_add['symbol'] = __symbol
                    __message_add['interval'] = __interval
                    __message_add['last_update_id'] = __temp_data['data']['time']
                    __message_add['open_time'] = int(__temp_data['data']['candles'][0]) * 1000
                    __message_add['close_time'] = (
                        (int(__temp_data['data']['candles'][0]) * 1000) + __delta_time
                    )
                    __message_add['open_time_date'] = (
                        time.strftime("%Y-%m-%d %H:%M:%S",\
                                        time.gmtime(int(round(__message_add['open_time']/1000))))
                    )
                    __message_add['close_time_date'] = (
                        time.strftime("%Y-%m-%d %H:%M:%S",\
                                        time.gmtime(int(round(__message_add['close_time']/1000))))
                    )
                    __message_add['open'] = __temp_data['data']['candles'][1]
                    __message_add['close'] = __temp_data['data']['candles'][2]
                    __message_add['hight'] = __temp_data['data']['candles'][3]
                    __message_add['low'] = __temp_data['data']['candles'][4]
                    __message_add['volume'] = __temp_data['data']['candles'][5]
                    __message_add['is_closed'] = None

                    self.__ws_temp_data[__stream_index][int(__message_add['open_time'])] = (
                        __message_add
                    )

                    while len(self.__ws_temp_data[__stream_index]) > self.__data_max_len:
                        __first_key = min(list(self.__ws_temp_data[__stream_index].keys()))
                        __nc = self.__ws_temp_data[__stream_index].pop(__first_key,None)

                    __message_out = list(self.__ws_temp_data[__stream_index].values())

                    result = __message_out[:self.__result_max_len]

        return result

    def manage_websocket_message_trades(self,data):
        """
        manage_websocket_message_trades
        ===============================
            This function manage websocket message and normalize result
            data for trades endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        if __temp_data is not None and isinstance(__temp_data, dict):
            if 'type' in __temp_data and __temp_data['type'] == 'message'\
                and  'data' in __temp_data and isinstance(__temp_data['data'], dict)\
                and 'symbol' in __temp_data['data']:

                __symbol = __temp_data['data']['symbol']
                __stream_index = self.get_stream_index('trades', __symbol)

                if self.__ws_temp_data[__stream_index] is None:
                    self.__ws_temp_data[__stream_index] = queue.Queue(maxsize=self.__data_max_len)

                __message_add = None
                __message_add = {}
                __message_add['endpoint'] = 'trades'
                __message_add['exchange'] = self.__exchange
                __message_add['symbol'] = __symbol
                __message_add['interval'] = None
                __message_add['event_time'] = int(round(time.time_ns() / 1000000))
                __message_add['trade_id'] = str(__temp_data['data']['tradeId'])
                __message_add['price'] = str(__temp_data['data']['price'])
                __message_add['quantity'] = str(__temp_data['data']['size'])
                __message_add['trade_time'] = (
                    int(round(float(__temp_data['data']['time']) / 1000000))
                )
                __message_add['trade_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",\
                                    time.gmtime(int(round(__message_add['trade_time']\
                                                        /1000))))\
                        + '.' + str(round(math.modf(round(__message_add['trade_time']/1000,\
                                                            3))[0]*1000)).rjust(3,'0')
                )
                __message_add['side_of_taker'] = __temp_data['data']['side'].upper()

                if self.__ws_temp_data[__stream_index].full():
                    self.__ws_temp_data[__stream_index].get(True,1)

                self.__ws_temp_data[__stream_index].put(__message_add,True,5)

                __message_out = list(self.__ws_temp_data[__stream_index].queue)
                ##__message_out.reverse()

                result = __message_out[:self.__result_max_len]

        return result

    def manage_websocket_message_ticker(self, data):
        """
        manage_websocket_message_ticker
        ===============================
            This function manage websocket message and normalize result
            data for ticker endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        if __temp_data is not None and isinstance(__temp_data,dict):
            if 'type' in __temp_data and __temp_data['type'] == 'message'\
                and  'data' in __temp_data and isinstance(__temp_data['data'],dict):

                __symbol = __temp_data['topic'].split(':')[1]
                __stream_index = self.get_stream_index('ticker', __symbol)

                self.__ws_temp_data[__stream_index] = __temp_data

                __message_add = None
                __message_add = {}
                __message_add['endpoint'] = 'ticker'
                __message_add['exchange'] = self.__exchange
                __message_add['symbol'] = __symbol
                __message_add['interval'] = None
                __message_add['event_type'] = '24hrTicker'
                __message_add['event_time'] = int(round(time.time_ns() / 1000000))
                __message_add['event_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",\
                                    time.gmtime(int(round(__message_add['event_time']\
                                                        /1000))))\
                        + '.' + str(round(math.modf(round(__message_add['event_time']\
                                                            /1000,3))[0]*1000)).rjust(3,'0')
                )
                __message_add['price_change'] = None
                __message_add['price_change_percent'] = None
                __message_add['weighted_average_price'] = None
                __message_add['first_trade_before_the_24hr_rolling_window'] = None
                __message_add['last_price'] = __temp_data['data']['price']
                __message_add['last_quantity'] = __temp_data['data']['size']
                __message_add['best_bid_price'] = __temp_data['data']['bestBid']
                __message_add['best_bid_quantity'] = __temp_data['data']['bestBidSize']
                __message_add['best_ask_price'] = __temp_data['data']['bestAsk']
                __message_add['best_ask_quantity'] = __temp_data['data']['bestAskSize']
                __message_add['open_price'] = None
                __message_add['high_price'] = None
                __message_add['low_price'] = None
                __message_add['total_traded_base_asset_volume'] = None
                __message_add['total_traded_quote_asset_volume'] = None
                __message_add['statistics_open_time'] = None
                __message_add['statistics_open_time_date'] = None
                __message_add['statistics_close_time'] = None
                __message_add['statistics_close_time_date'] = None
                __message_add['total_number_of_trades'] = None

                __message_out = __message_add

                result = __message_out

        return result

    def manage_websocket_message(self,ws,message_in): # pylint: disable=unused-argument
        """
        manage_websocket_message
        ========================
            This function manage websocket message and normalize result data.

                :param ws: WebSocketApp instance.
                :param message_in: str message from websocket.
                
                :return dict: Return dict with normalized data.
        """

        result = None

        try:
            if ccf.is_json(message_in):
                __temp_data = json.loads(message_in)
                __proc_data = False

                __endpoint = 'NONE'

                if __temp_data is not None\
                    and isinstance(__temp_data, dict)\
                    and 'topic' in __temp_data\
                    and __temp_data['topic'] is not None\
                    and isinstance(__temp_data['topic'], str)\
                    and len(__temp_data['topic']) > 0:

                    __tmp_endpoint = __temp_data['topic'].split(':')[0]

                    if __tmp_endpoint == '/market/level2':
                        __endpoint = 'order_book'
                    elif __tmp_endpoint == '/market/candles':
                        __endpoint = 'kline'
                    elif __tmp_endpoint == '/market/match':
                        __endpoint = 'trades'
                    elif __tmp_endpoint == '/market/ticker':
                        __endpoint = 'ticker'

                if __endpoint == 'order_book':
                    __message_out = self.manage_websocket_message_order_book(__temp_data)

                    result = {}
                    result['data'] = __message_out
                    result['min_proc_time_ms'] = 0
                    result['max_proc_time_ms'] = 0

                elif __endpoint == 'kline':
                    __message_out = self.manage_websocket_message_kline(__temp_data)

                    result = {}
                    result['data'] = __message_out
                    result['min_proc_time_ms'] = 0
                    result['max_proc_time_ms'] = 0

                elif __endpoint == 'trades':
                    __message_out = self.manage_websocket_message_trades(__temp_data)

                    result = {}
                    result['data'] = __message_out
                    result['min_proc_time_ms'] = 0
                    result['max_proc_time_ms'] = 0

                elif __endpoint == 'ticker':
                    __message_out = self.manage_websocket_message_ticker(__temp_data)

                    result = {}
                    result['data'] = __message_out
                    result['min_proc_time_ms'] = 0
                    result['max_proc_time_ms'] = 0

        except Exception as exc: # pylint: disable=broad-except
            print(str(exc))

        return result

    def __manage_websocket_diff_data(self, diff_data):

        result = False

        __symbol = None
        __stream_index = None

        current_data = None
        current_data_seq = 0

        if 'symbol' in diff_data\
            and diff_data['symbol'] is not None\
            and isinstance(diff_data['symbol'], str):
            __symbol = diff_data['symbol']
            __stream_index = self.get_stream_index('order_book', __symbol)

            current_data = self.__ws_temp_data[__stream_index]
            current_data_seq = current_data['last_update_id']

        if isinstance(diff_data['changes'],dict) and 'bids' in diff_data['changes']\
            and 'asks' in diff_data['changes'] and isinstance(diff_data['changes']['bids'],list)\
            and isinstance(diff_data['changes']['asks'],list) and __symbol is not None\
            and __stream_index is not None and current_data is not None:
            __temp_bids_d = {}
            __temp_asks_d = {}
            __temp_bids_l = []
            __temp_asks_l = []

            __key = 'bids'
            for i in range(0,len(current_data[__key])):
                __temp_bids_d[current_data[__key][i][0]] = current_data[__key][i][1]

            for i in range(0,len(diff_data['changes'][__key])):
                if int(diff_data['changes'][__key][i][2]) > current_data_seq:
                    if float(diff_data['changes'][__key][i][1]) == 0:
                        n_t = __temp_bids_d.pop(diff_data['changes'][__key][i][0],None) # pylint: disable=unused-variable
                    else:
                        __temp_bids_d[diff_data['changes'][__key][i][0]] = \
                            diff_data['changes'][__key][i][1]

            for i,j in __temp_bids_d.items():
                __temp_bids_l.append([i,j])

            __temp_bids = sorted(__temp_bids_l, key=lambda value: float(value[0]), reverse=True)

            __key = 'asks'
            for i in range(0,len(current_data[__key])):
                __temp_asks_d[current_data[__key][i][0]] = current_data[__key][i][1]

            for i in range(0,len(diff_data['changes'][__key])):
                if int(diff_data['changes'][__key][i][2]) > current_data_seq:
                    if float(diff_data['changes'][__key][i][1]) == 0:
                        n_t = __temp_asks_d.pop(diff_data['changes'][__key][i][0],None)
                    else:
                        __temp_asks_d[diff_data['changes'][__key][i][0]] = \
                            diff_data['changes'][__key][i][1]

            for i,j in __temp_asks_d.items():
                __temp_asks_l.append([i,j])

            __temp_asks = sorted(__temp_asks_l, key=lambda value: float(value[0]), reverse=False)

            current_data['bids'] = __temp_bids
            current_data['asks'] = __temp_asks
            current_data['last_update_id'] = int(diff_data['sequenceEnd'])
            current_data['diff_update_id'] = current_data['last_update_id'] - current_data_seq
            current_data['type'] = 'update'
            self.__ws_temp_data[__stream_index] = current_data
            result = True

        return result



"""
CCXW - CryptoCurrency eXchange Websocket Library
Main Class Ccxw

Author: Ricardo Marcelo Alvarez
Date: 2023-10-31
"""

import threading
import os
import os.path
import time
import datetime
import json
import sqlite3
import base64
import random
import gzip
import humanize

import websocket
from .binance import BinanceCcxwAuxClass
from .bybit import BybitCcxwAuxClass
from .bingx import BingxCcxwAuxClass
from .kucoin import KucoinCcxwAuxClass
from .okx import OkxCcxwAuxClass

class CcxwExchangeConfig:
    """
    A configuration class for managing exchange classes.

    Attributes:
    exchange_classes (dict): A dictionary mapping exchange names to their corresponding classes.

    Example:
    To obtain the class for the 'binance' exchange, use ExchangeConfig.exchange_classes['binance'].
    """

    exchange_classes = {
        'binance': BinanceCcxwAuxClass,
        'bybit': BybitCcxwAuxClass,
        'bingx': BingxCcxwAuxClass,
        'kucoin': KucoinCcxwAuxClass,
        'okx': OkxCcxwAuxClass
    }

class Ccxw():
    """
    CCXW - CryptoCurrency eXchange Websocket Library
    ================================================
    This class retrieves data from exchange WebSocket APIs and inserts or updates data
    in a temporary SQLite database, which is done in the background, making the data
    available for retrieval.
    Currently, it is available for the following exchanges: Binance, Bybit, etc.
    For a complete list, see Ccxw.get_supported_exchanges().
    It supports only the following endpoints: order_book, kline, etc.
    For a complete list, see Ccxw.get_supported_endpoints().

    Example:

    ```python
    import time
    import pprint
    from ccxw import Ccxw

    symbol = 'BTC/USDT'

    streams = [\
                    {\
                        'endpoint': 'order_book',\
                        'symbol': symbol
                    },\
                    {\
                        'endpoint': 'kline',\
                        'symbol': symbol,\
                        'interval': interval\
                    },\
                    {\
                        'endpoint': 'trades',\
                        'symbol': symbol
                    },\
                    {\
                        'endpoint': 'ticker',\
                        'symbol': symbol
                    }\
            ]
    
    wsm = Ccxw('binance',\
                streams,\
                result_max_len=5,\
                data_max_len=10)

    wsm.start()  # Start getting data

    time.sleep(2)  # Wait for available data

    for i in range(0, 10):
        for stream in streams:
            interval = 'none'
            if 'interval' in stream:
                interval = stream['interval']
            data = wsm.get_current_data(stream['endpoint'], stream['symbol'], interval)
            pprint.pprint(data, sort_dicts=False)
            print('----------------------------------')
            time.sleep(1)
        print('============================================================')

    wsm.stop()  # Stop getting data
    ```
    """

    def __init__(self, exchange, streams=list[dict], trading_type: str='SPOT',\
        testmode: bool=False, result_max_len: int=5,\
        data_max_len: int=400, debug: bool=False):
        """
        Ccxw constructor
        ================
            :param self: Ccxw instance.
            :param exchange: str exchange name.
            :param streams: list[dict]
                                    dicts must have this struct.
                                        {
                                            'endpoint': str only allowed 'order_book' | 'kline' |\
                                                'trades' | 'ticker',
                                            'symbol': str unified symbol.,
                                            'interval': str '1m' | '3m' | '5m' | '15m' | '30m' |\
                                                '1h' | '2h' | '4h' | '6h' | '8h' | '12h' | '1d' |\
                                                '3d' | '1w' | '1mo' for 'kline' endpoint is\
                                                    mandatory.
                                        }
            :param trading_type: str only allowed 'SPOT'.
            :param testmode: bool.
            :param result_max_len: int Max return values > 1 and <= data_max_len.
            :param data_max_len: int. > 1 and <= 400 max len of data getting from exchange.
            :param debug: bool Output verbosity.

            :return: Return a new instance of the Class Ccxw.
        """

        self.__debug = debug
        self.__data_max_len = data_max_len

        self.__data_max_len = min(self.__data_max_len, 400)
        self.__data_max_len = max(self.__data_max_len, 1)

        self.__current_ccxw_thread = threading.current_thread()

        if self.__current_ccxw_thread is not None:
            self.__current_ccxw_thread.name = 'class_ccxw_thread'

        if exchange not in Ccxw.get_supported_exchanges():
            raise ValueError('The exchange ' + str(exchange) + ' is not supported.')

        self.__key_sel = {}

        self.__exchange = None
        self.__ws = None
        self.__ws_url = None
        self.__ws_endpoint_url = None
        self.__ws_endpoint_on_auth_vars = None
        self.__ws_endpoint_on_open_vars = None
        self.__ws_endpoint_on_close_vars = None
        self.__ws_ping_interval = 0
        self.__ws_ping_timeout = None
        self.__socket = None
        self.__thread = None
        self.__stop_launcher = False # Used in methods self.start() and self.stop()
        self.__conn_db = None
        self.__cursor_db = None
        self.min_proc_time_ms = None
        self.max_proc_time_ms = 0
        self.__trading_type = 'SPOT'

        self.__result_max_len = result_max_len
        self.__ws_ended = True

        self.__ws_streams = streams

        if exchange in self.get_supported_exchanges():
            self.__exchange = exchange
            self.__testmode = testmode

            self.__trading_type = trading_type

            self.__auxiliary_class = CcxwExchangeConfig.exchange_classes[self.__exchange]\
                (streams=self.__ws_streams, trading_type=self.__trading_type,\
                testmode=self.__testmode,\
                result_max_len=self.__result_max_len,\
                data_max_len=self.__data_max_len, debug=self.__debug)

            self.__init_key_selector()

            self.__database_name = '/tmp/temp_database_'\
                + str(datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f"))\
                + '_' + str(random.randint(90000,99999)) + '_'\
                + str(random.randint(90000,99999)) + '.db'

            self.__table_name = 'temp_table'

            #self.__conn_db = sqlite3.connect(self.__database_name, check_same_thread=False)
            self.__conn_db = sqlite3.connect(':memory:', check_same_thread=False)
            self.__conn_db_lock = threading.Lock()

            self.__result_max_len = min(self.__result_max_len, self.__data_max_len)
            self.__result_max_len = max(self.__result_max_len, 1)

            self.__ws_url = self.__auxiliary_class.get_websocket_url()

            if isinstance(self.__ws_url,str) and len(self.__ws_url) > 0:

                __get_websocket_endpoint_data = (
                    self.__auxiliary_class.get_websocket_endpoint_path()
                )

                if __get_websocket_endpoint_data is not None\
                    and isinstance(__get_websocket_endpoint_data,dict):
                    if 'ws_endpoint_url' in __get_websocket_endpoint_data:
                        self.__ws_endpoint_url = (
                            __get_websocket_endpoint_data['ws_endpoint_url']
                        )

                    if 'ws_endpoint_on_open_vars' in __get_websocket_endpoint_data:
                        self.__ws_endpoint_on_open_vars = (
                            __get_websocket_endpoint_data['ws_endpoint_on_open_vars']
                        )

                    if 'ws_endpoint_on_close_vars' in __get_websocket_endpoint_data:
                        self.__ws_endpoint_on_close_vars = (
                            __get_websocket_endpoint_data['ws_endpoint_on_close_vars']
                        )

                    self.__socket = self.__ws_url + self.__ws_endpoint_url

                    if 'ws_ping_interval' in __get_websocket_endpoint_data:
                        self.__ws_ping_interval = (
                            __get_websocket_endpoint_data['ws_ping_interval']
                        )

                    if 'ws_ping_timeout' in __get_websocket_endpoint_data:
                        self.__ws_ping_timeout = (
                            __get_websocket_endpoint_data['ws_ping_timeout']
                        )

                    if 'ws_endpoint_on_auth_vars' in __get_websocket_endpoint_data:
                        self.__ws_endpoint_on_auth_vars = (
                            __get_websocket_endpoint_data['ws_endpoint_on_auth_vars']
                        )
            else:
                raise ValueError('The exchange ' + str(exchange) + ' have not websocket api.')

        else:
            raise ValueError('The exchange ' + str(exchange) + ' have not websocket api.')

    def __del__(self):

        if not self.__stop_launcher:
            self.stop()

        # if self.__auxiliary_class is not None and hasattr(self.__auxiliary_class, 'stop'):
        #     self.__auxiliary_class.stop()

        if hasattr(self,'__conn_db') and self.__conn_db is not None:
            with self.__conn_db_lock:
                self.__conn_db.commit()
                self.__conn_db.close()

        if hasattr(self,'__database_name') and self.__database_name is not None:
            if os.path.exists(self.__database_name):
                os.remove(self.__database_name)

        del self.__auxiliary_class

    def __init_key_selector(self):
        result = False

        for stream in self.__ws_streams:
            __index = None
            __interval = 'none'
            if 'interval' in stream:
                __interval = stream['interval']

            __index = self.__auxiliary_class.get_stream_index(stream['endpoint'],\
                                                              stream['symbol'],\
                                                              __interval)

            self.__key_sel[__index] = __index
        return result

    def get_exchange_info(self):
        """
        get_exchange_info
        =================
            This function get exchange info. 
                :return dict: Return exchange info.
        """

        result = None

        result = self.__auxiliary_class.get_exchange_info()

        return result

    def get_exchange_full_list_symbols(self, sort_list=True):
        """
        get_exchange_full_list_symbols
        ==============================
            This function get exchange info. 
                :param sort_list: bool.
                :return dict: Return exchange info.
        """

        result = None

        result = self.__auxiliary_class.get_exchange_full_list_symbols(sort_list)

        return result

    def start(self):
        """
        Starting getting data form exchange
        ===================================
            This method start a websocket client in a new thread.
                :param self: Ccxw instance.

                :return bool: Return True if starting OK or False 
        """
        result = False

        if self.__auxiliary_class is not None and hasattr(self.__auxiliary_class, 'start'):
            self.__auxiliary_class.start()
            time.sleep(2)

        try:
            self.__stop_launcher = False # Used in self.stop()
            self.__thread = threading.Thread(target=self.__websocket_launcher,\
                                             args=(self.__socket,),\
                                             name='ccxw_websocket_thread')
            self.__thread.start()
            result = True
        except Exception: # pylint: disable=broad-except
            result = False

        time.sleep(5)

        return result

    def stop(self,time_to_wait=5):
        """
        Stopping getting data form exchange
        ===================================
            This method stop a websocket client and join and wait the thread ended.
                :param self: Ccxw instance.
                :param time_to_wait: int Max time to wait the websocket ended.

                :return bool: Return True if starting OK or False 
        """
        result = False

        __time_counter = 0
        __time_limit = 90

        if self.__ws and self.__ws_endpoint_on_close_vars is not None:
            if isinstance(self.__ws_endpoint_on_close_vars, str)\
                and len(self.__ws_endpoint_on_close_vars) > 0:
                self.__ws.send(self.__ws_endpoint_on_close_vars)
            if isinstance(self.__ws_endpoint_on_close_vars, list)\
                and len(self.__ws_endpoint_on_close_vars) > 0:
                for __close_vars in self.__ws_endpoint_on_close_vars:
                    self.__ws.send(__close_vars)

        time.sleep(2)

        if self.__auxiliary_class is not None and hasattr(self.__auxiliary_class, 'stop'):
            self.__auxiliary_class.stop()

        self.__stop_launcher = True

        if not self.__ws_ended:
            try:
                self.__ws.close()
                result = True
            except Exception: # pylint: disable=broad-except
                result = False

        while not self.__ws_ended and __time_counter <= __time_limit:
            __time_counter = __time_counter + 1
            time.sleep(1)

        if not self.__ws_ended:
            try:
                self.__ws.close()
                result = True
            except Exception: # pylint: disable=broad-except
                result = False
        else:
            result = True

        if self.__thread is not None:
            self.__thread.join(time_to_wait)

        # print('ENTRE: 8888')
        # pprint.pprint(str(self.__auxiliary_class))

        return result

    def __websocket_launcher(self, socket):
        """
        Initialize temporal database and starting getting data from websocket
        ======================================================================
            This is a Class internal method and running in a new thread,
            First initialize a temporal database and then initialize and launch websocket client
                :param self: Ccxw instance.
                :param str: websocket URL.

                :return bool: Return True if starting OK or False 
        """
        result = False

        try:
            with self.__conn_db_lock:
                self.__cursor_db = self.__conn_db.cursor()

                __sql_create_table_db = f'DROP TABLE IF EXISTS {self.__table_name};\n'
                __sql_create_table_db += f'CREATE TABLE IF NOT EXISTS {self.__table_name} \
                                            (key_data VARCHAR(40) PRIMARY KEY, value_data TEXT);\n'

                self.__cursor_db.executescript(__sql_create_table_db)

                for __key_data in self.__key_sel.values():
                    __sql_insert_data = f'INSERT INTO {self.__table_name} \
                                            (key_data, value_data) VALUES (?, NULL);\n'
                    self.__cursor_db.execute(__sql_insert_data, (str(__key_data),))

                self.__conn_db.commit()
                result = True
        except Exception as exc: # pylint: disable=broad-except
            result = False
            print(str(exc))

        websocket.enableTrace(self.__debug)

        try:
            self.__ws = (
                websocket.WebSocketApp(socket, on_message=self.__manage_websocket_message,\
                                       on_open=self.__manage_websocket_open,\
                                       on_close=self.__manage_websocket_close)
            )
            result = True

        except Exception as exc: # pylint: disable=broad-except
            result = False
            print('On create websocket exception: ' + str(exc))
        self.__ws_ended = False
        __ws_temp = (
            self.__ws.run_forever(ping_interval=self.__ws_ping_interval,\
                                  ping_timeout=self.__ws_ping_timeout, reconnect=320)
            )

        self.__ws_ended = True

        return result

    def __manage_websocket_open(self, ws):
        """
        websocket.WebSocketApp on_open function.
        ========================================
            Some exchange endpoints require send specific data when websocket is opened 
                :param self: Ccxw instance.
                :param ws: websocket.WebSocketApp instance.

                :return None:
        """
        result = False

        if self.__ws_endpoint_on_auth_vars is not None\
            and isinstance(self.__ws_endpoint_on_auth_vars, str):
            try:
                ws.send(self.__ws_endpoint_on_auth_vars)
                result = True
            except Exception: # pylint: disable=broad-except
                result = False

        elif self.__ws_endpoint_on_open_vars is not None:
            if isinstance(self.__ws_endpoint_on_open_vars, str):
                try:
                    ##print('OV: ' + self.__ws_endpoint_on_open_vars)
                    ws.send(self.__ws_endpoint_on_open_vars)
                    result = True
                except Exception: # pylint: disable=broad-except
                    result = False
            elif isinstance(self.__ws_endpoint_on_open_vars, list):
                try:
                    for __send_data in self.__ws_endpoint_on_open_vars:
                        ##print('OV: ' + str(__send_data))
                        ws.send(__send_data)
                        time.sleep(0.14)
                    result = True
                except Exception: # pylint: disable=broad-except
                    result = False

        else:
            result = True

        return result

    def __manage_websocket_close(self,ws,close_status_code,close_msg):
        """
        websocket.WebSocketApp on_close function.
        =========================================
            Some exchange endpoints require send specific data when websocket is closed
                :param self: Ccxw instance.
                :param ws: websocket.WebSocketApp instance.
                :param close_status_code: int close status code.
                :param close_msg: str close message.

                :return None:
        """

        __tmp_code = close_status_code
        __tmp_msg = close_msg

        if ws is not None:
            self.__ws_ended = True
            time.sleep(5)


    def ws_on_error(self, ws,error):
        """
        websocket.WebSocketApp on_error function.
        =========================================
            :param self: Ccxw instance.
            :param ws: websocket.WebSocketApp instance.
            :param error: Exception object.

            :return None: 
        """
        err_msg = "Found error in websocket connection "\
            + str(ws.url)  + " -> WS: " + str(ws)\
            + ", error: " + str(error)
        print(err_msg)

    def __manage_websocket_message(self, ws, message_in):
        """
        websocket.WebSocketApp on_message function.
        ===========================================
            This function process the data. The data is converted in json string,
            compress, encode in base64 and then insert or update in temporal database
                :param self: Ccxw instance.
                :param ws: websocket.WebSocketApp instance.
                :param message_in: str message.

                :return None:
        """

        __time_ini = time.time_ns()

        #print(str(message_in))

        try:
            __managed_data = self.__auxiliary_class.manage_websocket_message(ws, message_in)

            #pprint.pprint(__managed_data, sort_dicts=False)

            if __managed_data is not None and isinstance(__managed_data,dict)\
                and 'data' in __managed_data and 'max_proc_time_ms' in __managed_data:

                __endpoint = ''
                __symbol = ''
                __interval = 'none'

                if __managed_data['data'] is not None:
                    if isinstance(__managed_data['data'], list):
                        if len(__managed_data['data']) > 0:
                            __endpoint = __managed_data['data'][0]['endpoint']
                            __symbol = __managed_data['data'][0]['symbol']

                            if 'interval' in __managed_data['data'][0]\
                                and __managed_data['data'][0]['interval'] is not None:
                                __interval = __managed_data['data'][0]['interval']

                    elif isinstance(__managed_data['data'], dict):
                        __endpoint = __managed_data['data']['endpoint']
                        __symbol = __managed_data['data']['symbol']

                        if 'interval' in __managed_data['data']\
                            and __managed_data['data']['interval'] is not None:
                            __interval = __managed_data['data']['interval']

                #print('I: ' + str(__endpoint) + ', ' + str(__symbol) + ', ' + str(__symbol))

                if len(__endpoint) > 0 and len(__symbol) > 0:
                    __index_key_sel = self.__auxiliary_class.get_stream_index(__endpoint,\
                                                                            __symbol,\
                                                                            __interval)

                    __ws_temp_data = __managed_data
                    __ws_temp_data['min_proc_time_ms'] = self.min_proc_time_ms
                    __ws_temp_data['max_proc_time_ms'] = self.max_proc_time_ms
                    __message_out = json.dumps(__ws_temp_data)
                    __message_out = gzip.compress(bytes(__message_out,'utf-8'), compresslevel=9)
                    __message_base64 = base64.b64encode(__message_out).decode("ascii")
                    __sql_update = (
                        f'UPDATE {self.__table_name} SET value_data = ? WHERE key_data = ?;'
                    )

                    with self.__conn_db_lock:
                        self.__cursor_db = self.__conn_db.cursor()
                        self.__cursor_db.execute(__sql_update,\
                                                (str(__message_base64),\
                                                str(self.__key_sel[__index_key_sel])))
                        self.__conn_db.commit()

        except Exception as exc: # pylint: disable=broad-except
            print(str(exc))

        __time_diff = time.time_ns() - __time_ini
        __time_diff = round((__time_diff / 1000000),3)

        if __time_diff > self.max_proc_time_ms:
            self.max_proc_time_ms = __time_diff

        if self.min_proc_time_ms is None or __time_diff < self.min_proc_time_ms:
            self.min_proc_time_ms = __time_diff

        if self.__stop_launcher and not self.__ws_ended:
            ws.close()

    def get_current_data(self, endpoint, symbol, interval='none'):
        """
        Ccxw get_current_data function.
        ===============================
            This function get de data from temporal database decode from base64,
            decompress and convert json data in dict and then return this data
                :param self: Ccxw instance.
                :param endpoint: str.
                :param symbol: str.
                :param interval: str.

                :return: dict with last data.
        """
        result = None

        __index_key_sel = self.__auxiliary_class.get_stream_index(endpoint, symbol, interval)

        __sql_select = f'SELECT value_data FROM "{self.__table_name}" WHERE key_data = ?;'

        try:
            with self.__conn_db_lock:
                __local_cursor_db = self.__conn_db.cursor()
                __local_cursor_db.execute(__sql_select, (str(self.__key_sel[__index_key_sel]),))
                __current_data = __local_cursor_db.fetchone()

            if __current_data is not None\
                and isinstance(__current_data,(list, tuple))\
                and len(__current_data) > 0 and __current_data[0] is not None\
                and isinstance(__current_data[0],str):
                result = (
                    json.loads(gzip.decompress(base64.b64decode(__current_data[0])).decode('utf-8'))
                )

                if result is not None and isinstance(result, dict) and 'data' in result\
                    and result['data'] is not None:
                    __data = result['data']
                    if isinstance(__data, dict):
                        if 'symbol' in __data:
                            __data['symbol'] = self.__auxiliary_class\
                                .get_unified_symbol_from_symbol(__data['symbol'])

                        if 'interval' in __data:
                            __data['interval'] = self.__auxiliary_class\
                                .get_unified_interval_from_interval(__data['interval'])

                    elif isinstance(__data, list):
                        for res in __data:
                            if isinstance(res, dict):
                                if 'symbol' in res:
                                    res['symbol'] = self.__auxiliary_class\
                                        .get_unified_symbol_from_symbol(res['symbol'])

                                if 'interval' in res:
                                    res['interval'] = self.__auxiliary_class\
                                        .get_unified_interval_from_interval(res['interval'])

                    result['data'] = __data

        except Exception as exc: # pylint: disable=broad-except
            print(str(exc))

        return result

    def get_sqlite_memory_used(self):
        """
        Ccxw get_sqlite_memory_used function.
        =====================================
            This method return a estimated bytes used for sqlite temporary database
                :param self: Ccxw instance.

                :return: int.
        """

        result = 0

        with self.__conn_db_lock:
            try:
                __sql_to_exec = 'select page_size * page_count'
                __sql_to_exec += ' from pragma_page_count(), pragma_page_size();'
                cursor = self.__conn_db.cursor()
                cursor.execute(__sql_to_exec)
                result = int(cursor.fetchone()[0])
            except Exception: # pylint: disable=broad-except
                result = 0

        return result

    def get_sqlite_memory_used_human_readable(self):
        """
        Ccxw get_sqlite_memory_used_human_readable function.
        ====================================================
            This method return a estimated bytes used for sqlite temporary database\
                ( In human readable format)
                :param self: Ccxw instance.

                :return: str.
        """

        result = ''
        __db_size = self.get_sqlite_memory_used()
        result = humanize.naturalsize(__db_size)

        return result


    @classmethod
    def get_supported_exchanges(cls):
        """
        Ccxw get_supported_exchanges function.
        ======================================
            This method return a list of supported exchanges.
                :param cls: Ccxw Class.

                :return: list of supported exchanges.
        """
        __suported_exchanges = ['binance' , 'bybit', 'kucoin', 'bingx','okx']

        return __suported_exchanges

    @classmethod
    def get_supported_endpoints(cls):
        """
        Ccxw get_supported_endpoints function.
        ======================================
            This method return a list of supported end points.
                :param cls: Ccxw Class.

                :return: list of supported endpoints.
        """
        __suported_endpoints = ['order_book', 'kline', 'trades','ticker']

        return __suported_endpoints



# pylint: disable=too-many-lines
"""
Ccxw - CryptoCurrency eXchange Websocket Library
Bingx auxiliary functions

Author: Ricardo Marcelo Alvarez
Date: 2023-10-31
"""

import os
import json
import time
import datetime
import gzip
import io
import math
import random
import threading
import websocket
import websocket_server

import ccxw.ccxw_common_functions as ccf
from ccxw.safe_thread_vars import DictSafeThread
import ccxw

class BingxCcxwAuxClass():
    """
    Ccxw - CryptoCurrency eXchange Websocket Library BingxCcxwAuxClass
    ==================================================================
        This class contains helper functions for the Ccxw class.
    """

    def __init__(self,\
                 streams: list[dict],\
                 trading_type: str='SPOT',\
                 testmode: bool=False,\
                 result_max_len: int=5,
                 data_max_len: int=1000,\
                 debug: bool=False):
        """
        BingxCcxwAuxClass constructor
        =============================
            :param self: BingxCcxwAuxClass instance.
            :param streams: list[dict]
                                    dicts must have this struct.
                                        {
                                            'endpoint': str only allowed 'order_book' | 'kline' |\
                                                'trades' | 'ticker',
                                            'symbol': str unified symbol.,
                                            'interval': str '1m' for 'kline' endpoint is\
                                                    mandatory.
                                        }

            :param trading_type: str only allowed 'SPOT'.
            :param testmode: bool.
            :param result_max_len: int Max return values > 1 and <= data_max_len.
            :param data_max_len: int. > 1 and <= 400 max len of data getting from exchange.
            :param debug: bool Verbose output.

            :return: Return a new instance of the Class BingxCcxwAuxClass.
        """

        __exchange_limit_streams = 1024

        self.__exchange = os.path.basename(__file__)[:-3]
        self.__ws_streams = streams
        self.__ws_server = None
        self.__ws_server_url = '127.0.0.1'
        self.__ws_server_port_min = 10000
        self.__ws_server_port_max = 50000

        self.__ws_ping_interval = 0
        self.__ws_ping_timeout = None

        self.__ws_server = None
        self.__ws_server_url = '127.0.0.1'
        self.__ws_server_port_min = 10000
        self.__ws_server_port_max = 50000

        self.__ws_server_port = self.__ws_server_port_min

        self.__local_ws_url = 'ws://' + self.__ws_server_url + ':' + str(self.__ws_server_port)

        self.__testmode = testmode

        self.__exchange_info_cache = {}
        self.__exchange_info_cache['data'] = None
        self.__exchange_info_cache['last_get_time'] = 0

        self.__debug = debug
        self.__trading_type = trading_type
        self.__data_max_len = data_max_len

        self.__data_max_len = min(self.__data_max_len, 400)
        self.__data_max_len = max(self.__data_max_len, 1)

        self.__result_max_len = result_max_len

        self.__result_max_len = min(self.__result_max_len, 500)
        self.__result_max_len = max(self.__result_max_len, 1)

        self.__ws_url_api = None
        self.__ws_url_test = None
        self.__ws_url = None
        self.__url_api = None
        self.__url_test = None
        self.__ws_endpoint_url = None
        self.__ws_endpoint_on_open_vars = None
        self.__ws_endpoint_on_close_vars = None

        self.__ws_client = None
        self.__ws_endpoint_url_client = None
        self.__ws_endpoint_on_open_vars_client = None
        self.__ws_endpoint_on_close_vars_client = None

        self.__api_data_vars = None

        self.__ws_temp_data = DictSafeThread()
        self.__lock = threading.Lock()

        self.__thread_websocket_client = None
        self.__thread_api_client = None

        websocket.enableTrace(self.__debug)

        self.__is_stopped = False

        if not self.__check_streams_struct(streams):
            raise ValueError('The streams struct is not valid' + str(streams))

        if len(streams) > __exchange_limit_streams:
            raise ValueError('The exchange ' + str(self.__exchange)\
                             + ' not alowed more than ' + str(__exchange_limit_streams)\
                             + ' of streams.')

    def __check_streams_struct(self, streams):
        """
        __check_streams_struct
        ======================
            This function check streams struct
                :param self: This class instance.
                :param streams: list[dict]
                                dicts must have this struct.
                                    {
                                        'endpoint': str only allowed 'order_book' | 'kline' |\
                                            'trades' | 'ticker',
                                        'symbol': str unified symbol.,
                                        'interval': str '1m' for 'kline' endpoint is\
                                            mandatory.
                                    }

                :return bool: True if streams struct is correct otherwise False
        """
        result = False

        if streams is not None and isinstance(streams, list) and len(streams) > 0:
            result = True
            for stream in streams:
                result = result and stream is not None\
                    and isinstance(stream, dict)\
                    and 'endpoint' in stream\
                    and stream['endpoint'] is not None\
                    and isinstance(stream['endpoint'], str)\
                    and stream['endpoint'] in ccxw.Ccxw.get_supported_endpoints()\
                    and 'symbol' in stream\
                    and stream['symbol'] is not None\
                    and isinstance(stream['symbol'], str)\
                    and self.if_symbol_supported(stream['symbol'])

                if result and stream['endpoint'] == 'kline':
                    result = 'interval' in stream\
                        and stream['interval'] is not None\
                        and isinstance(stream['interval'], str)\
                        and stream['interval'] in\
                            ['1m']

        return result

    def __set_ws_server(self):
        result = False

        __host = '127.' + str(random.randint(0, 254)) + '.' + str(random.randint(0, 254))\
            + '.' + str(random.randint(1, 254))
        __port = random.randint(self.__ws_server_port_min, self.__ws_server_port_max)

        while not ccf.is_port_free(__port, __host):
            __host = '127.' + str(random.randint(0, 254)) + '.' + str(random.randint(0, 254))\
                + '.' + str(random.randint(1, 254))
            __port = random.randint(self.__ws_server_port_min, self.__ws_server_port_max)

        self.__ws_server_url = __host
        self.__ws_server_port = __port

        result = True

        return result

    def __start_ws_server(self):
        result = False
        __attemps = 0
        __attemps_limit = 900

        while not ccf.is_port_free(self.__ws_server_port, self.__ws_server_url)\
            and __attemps <= __attemps_limit:
            __attemps = __attemps + 1
            time.sleep(1)

        if ccf.is_port_free(self.__ws_server_port, self.__ws_server_url):
            with self.__lock:
                self.__ws_server = websocket_server.WebsocketServer(host=self.__ws_server_url,\
                                                                    port=self.__ws_server_port)

                self.__ws_server.run_forever(threaded=True)
                result = True

        return result

    def __del__(self):
        with self.__lock:
            self.__is_stopped = True

        if self.__thread_websocket_client is not None\
            and threading.current_thread() is not self.__thread_websocket_client\
            and self.__thread_websocket_client.is_alive():
            try:
                self.__thread_websocket_client.join(9)
            except Exception: # pylint: disable=broad-except
                pass

        if self.__thread_api_client is not None\
            and threading.current_thread() is not self.__thread_api_client\
            and self.__thread_api_client.is_alive():
            try:
                self.__thread_api_client.join(9)
            except Exception: # pylint: disable=broad-except
                pass

    def start(self):
        """
        start
        =====

            :param self: OkxAuxClass instance.
        """
        with self.__lock:
            self.__is_stopped = False
        self.__start_ws_server()
        time.sleep(1)
        self.__start_ws_client()
        time.sleep(1)
        self.__start_get_data_from_api()

    def stop(self):
        """
        stop
        ====

            :param self: OkxAuxClass instance.
        """
        with self.__lock:
            self.__is_stopped = True
        self.__stop_ws_client()
        time.sleep(1)
        self.__stop_get_data_from_api()

        time.sleep(2)

        if self.__ws_server is not None:
            self.__ws_server.deny_new_connections()
            self.__ws_server.disconnect_clients_gracefully()
            time.sleep(1)
            self.__ws_server.shutdown_gracefully()
            self.__ws_server.disconnect_clients_gracefully()
            time.sleep(1)
            self.__ws_server.disconnect_clients_abruptly()
            #self.__ws_server.shutdown_abruptly()
            #self.__ws_server = None

    def __manage_websocket_open(self, ws):
        result = False

        if hasattr(ws, 'on_open_vars')\
            and ws.on_open_vars is not None\
            and isinstance(ws.on_open_vars, dict)\
            and all(key in ws.on_open_vars for key in ['id', 'reqType', 'dataType']):
            try:
                if ws.on_open_vars['id'] is not None and isinstance(ws.on_open_vars['id'], str)\
                    and ws.on_open_vars['reqType'] is not None\
                    and isinstance(ws.on_open_vars['reqType'], str)\
                    and ws.on_open_vars['dataType'] is not None:

                    if isinstance(ws.on_open_vars['dataType'], str):
                        __local_open_vars = None
                        __local_open_vars = json.dumps(ws.on_open_vars)
                        ws.send(__local_open_vars)
                        result = True
                    elif isinstance(ws.on_open_vars['dataType'], list):
                        for __local_data_type in ws.on_open_vars['dataType']:
                            __local_open_vars = None
                            __local_open_vars = {}
                            __local_open_vars['id'] = ws.on_open_vars['id']
                            __local_open_vars['reqType'] = ws.on_open_vars['reqType']
                            __local_open_vars['dataType'] = __local_data_type
                            __local_open_vars = json.dumps(__local_open_vars)

                            ws.send(__local_open_vars)
                            time.sleep(0.14)

                        result = True
            except Exception: # pylint: disable=broad-except
                result = False

        return result


    def __manage_websocket_close(self, ws, close_status_code, close_msg): # pylint: disable=unused-argument
        result = False

        if hasattr(ws, 'on_close_vars')\
            and ws.on_close_vars is not None\
            and isinstance(ws.on_close_vars, dict)\
            and all(key in ws.on_close_vars for key in ['id', 'reqType', 'dataType']):
            try:

                if ws.on_close_vars['id'] is not None and isinstance(ws.on_close_vars['id'], str)\
                    and ws.on_close_vars['reqType'] is not None\
                    and isinstance(ws.on_close_vars['reqType'], str)\
                    and ws.on_close_vars['dataType'] is not None:

                    if isinstance(ws.on_close_vars['dataType'], str):
                        __local_close_vars = None
                        __local_close_vars = json.dumps(ws.on_close_vars)
                        ws.send(__local_close_vars)
                        result = True
                    elif isinstance(ws.on_close_vars['dataType'], list):
                        for __local_data_type in ws.on_close_vars['dataType']:
                            __local_close_vars = None
                            __local_close_vars = {}
                            __local_close_vars['id'] = ws.on_close_vars['id']
                            __local_close_vars['reqType'] = ws.on_close_vars['reqType']
                            __local_close_vars['dataType'] = __local_data_type
                            __local_close_vars = json.dumps(__local_close_vars)
                            ws.send(__local_close_vars)
                            time.sleep(0.14)

                        result = True
            except Exception: # pylint: disable=broad-except
                result = False

        return result

    def __manage_websocket_message_local(self, ws, message): # pylint: disable=unused-argument

        if self.__ws_server is not None:
            if message is not None:
                if isinstance(message,bytes):
                    message = gzip.GzipFile(fileobj=io.BytesIO(message), mode='rb')
                    message = message.read()
                    message = message.decode('utf-8')

                    with self.__lock:
                        self.__ws_server.send_message_to_all(message)

    def __start_ws_client(self):
        result = True
        with self.__lock:
            self.__is_stopped = False
        self.__thread_api_client = threading.Thread(target=self.__start_ws)
        self.__thread_api_client.start()
        return result

    def __start_ws(self):
        result = False
        __socket = self.__ws_url + self.__ws_endpoint_url_client
        try:
            self.__ws_client = (
                websocket.WebSocketApp(__socket,\
                                       on_message=self.__manage_websocket_message_local,\
                                       on_open=self.__manage_websocket_open,\
                                       on_close=self.__manage_websocket_close)
            )

            self.__ws_client.on_open_vars = self.__ws_endpoint_on_open_vars_client
            self.__ws_client.on_close_vars = self.__ws_endpoint_on_close_vars_client

            result = True

            __ws_temp = (
                self.__ws_client.run_forever(ping_interval=self.__ws_ping_interval,\
                                             ping_timeout=self.__ws_ping_timeout,\
                                             reconnect=320)
                )

        except Exception as exc: # pylint: disable=broad-except
            result = False
            print('On create websocket exception: ' + str(exc))

        return result

    def __start_get_data_from_api(self):
        result = True
        with self.__lock:
            self.__is_stopped = False
        self.__thread_api_client = threading.Thread(target=self.__get_data_from_api)
        self.__thread_api_client.start()
        return result

    def __stop_get_data_from_api(self):
        with self.__lock:
            self.__is_stopped = True

        if self.__thread_api_client is not None\
            and threading.current_thread() is not self.__thread_api_client\
            and self.__thread_api_client.is_alive():

            try:
                self.__thread_api_client.join(9)
            except Exception: # pylint: disable=broad-except
                pass

    def __stop_ws_client(self):
        with self.__lock:
            self.__is_stopped = True

        self.__ws_client.close()

        if self.__thread_websocket_client is not None\
            and threading.current_thread() is not self.__thread_websocket_client\
            and self.__thread_websocket_client.is_alive():

            try:
                self.__thread_websocket_client.join(9)
            except Exception: # pylint: disable=broad-except
                pass

    def get_websocket_url(self):
        """
        get_websocket_url
        =================
            This function set and return websocket URL.

                :return str: Return websocket URL.
        """

        result = None

        if self.__set_ws_server():
            self.__local_ws_url = 'ws://' + self.__ws_server_url + ':' + str(self.__ws_server_port)

            if self.__trading_type == 'SPOT':
                self.__ws_url = 'wss://open-api-ws.bingx.com'
                self.__ws_url_api = self.__local_ws_url
                self.__ws_url_test = self.__ws_url

            result = self.__ws_url_api
            if self.__testmode:
                result = self.__ws_url_test

        return result

    def get_api_url(self):
        """
        get_api_url
        ===========
            This function set and return API URL.

                :return str: Return API URL.
        """

        result = None

        if self.__trading_type == 'SPOT':
            self.__url_api = 'https://open-api.bingx.com'
            self.__url_test = ''

        result = self.__url_api
        if self.__testmode:
            result = self.__url_test

        return result

    def get_exchange_info(self):
        """
        get_exchange_info
        =================
            This function get exchange info.

                :return dict: Return exchange info.
        """

        result = None

        max_last_get_time = 7200

        current_time = int(time.time())

        if (current_time - self.__exchange_info_cache['last_get_time']) >= max_last_get_time:
            self.__exchange_info_cache['data'] = None

        if self.__exchange_info_cache['data'] is None:

            __l_url_api = self.get_api_url()

            __l_endpoint = '/openApi/' + self.__trading_type.lower() + '/v1/common/symbols'

            __l_url_point = __l_url_api + __l_endpoint

            __data = ccf.file_get_contents_url(__l_url_point)

            if __data is not None and ccf.is_json(__data):
                result = json.loads(__data)
                self.__exchange_info_cache['data'] = result
                self.__exchange_info_cache['last_get_time'] = current_time

        else:
            result = self.__exchange_info_cache['data']

        return result

    def get_exchange_full_list_symbols(self, sort_list=True):
        """
        get_exchange_full_list_symbols
        ==============================
            This function get exchange info. 
                :param sort_list: bool.
                :return dict: Return exchange info.
        """

        result = None
        __main_data = self.get_exchange_info()

        if __main_data is not None and isinstance(__main_data,dict)\
            and 'data' in __main_data and isinstance(__main_data['data'],dict)\
            and 'symbols' in __main_data['data']:
            if isinstance(__main_data['data']['symbols'],list):

                result = []
                for symbol_data in __main_data['data']['symbols']:
                    if symbol_data is not None and isinstance(symbol_data,dict)\
                        and 'symbol' in symbol_data and isinstance(symbol_data['symbol'],str):
                        result.append(\
                                        symbol_data['symbol'].replace('-','/').upper())

                if sort_list:
                    result.sort()

        return result

    def get_unified_symbol_from_symbol(self, symbol):
        """
        get_unified_symbol_from_symbol
        ==============================
            This function get unified symbol from symbol.
                :param self: This class instance.
                :param symbol: str.
                :return str: Return unified symbol.
        """
        result = symbol

        full_list_symbols = self.get_exchange_full_list_symbols(False)

        for symbol_rpl in full_list_symbols:
            if symbol.replace('-','').replace('/', '').lower()\
                == symbol_rpl.replace('-','').replace('/', '').lower():
                result = symbol_rpl
                break

        return result

    def __get_interval_from_unified_interval(self, interval):
        result = '1'

        if interval is not None:
            result = str(interval)

            if 'm' in result:
                result = result.replace('m', 'min')
            elif result == '1h':
                result = '60m'
            elif 'h' in result:
                result = result.replace('h', 'hour')
            elif 'd' in result:
                result = result.replace('d', 'day')
            elif 'w' in result:
                result = result.replace('w', 'week')
            elif 'mo' in result:
                result = result.replace('mo', 'mon')

            result = str(result)

        return result

    def get_unified_interval_from_interval(self, interval):
        """
        get_unified_interval_from_interval
        ==================================
            This function get unified interval from interval.
                :param self: This class instance.
                :param interval: str.
                :return str: Return unified interval.
        """

        result = interval

        if result is not None:
            if result == '60m':
                result = '1h'
            elif 'min' in result:
                result = result.replace('min', 'm')
            elif 'hour' in result:
                result = result.replace('hour', 'h')
            elif 'day' in result:
                result = result.replace('day', 'd')
            elif 'week' in result:
                result = result.replace('week', 'w')
            elif 'mon' in result:
                result = result.replace('mon', 'mo')

            result = str(result)

        return result

    def if_symbol_supported(self, symbol):
        """
        if_symbol_supported
        ===================
            This function check if symbol is supported by the exchange.

                :return bool: Return True if supported 
        """

        result = False

        __data = self.get_exchange_full_list_symbols()

        if isinstance(__data,list) and symbol in __data:
            result = True

        return result

    def get_stream_index(self, endpoint, symbol, interval: str='none'):
        """
        get_stream_index
        ================
            This function return a stream index for use in dict.
                :param self: This class instance.
                :param endpoint: str.
                :param symbol: str unified symbol.
                :param interval: str.
                :return str: Return stream index
        """
        result = 'stream'

        if interval is None:
            interval = 'none'

        result = result + '_' + endpoint + '_' + symbol + '_' + interval

        result = result.replace('/','').replace('-', '').lower()

        return result

    def get_websocket_endpoint_path(self):
        """
        get_websocket_endpoint_path
        ===========================
            This function set endpoint URL and on_open vars and on_close vars.

                :return dict: Return dict with endpoint, open_vars and close_vars
        """

        result = None

        __send_data_vars = None

        self.__ws_endpoint_url = ''

        self.__ws_endpoint_url_client = '/market'
        self.__ws_endpoint_on_open_vars_client = None
        self.__ws_endpoint_on_close_vars_client = None
        self.__ws_endpoint_on_open_vars_client = {}
        self.__ws_endpoint_on_close_vars_client = {}
        self.__ws_endpoint_on_open_vars_client['id'] = str(time.time_ns())
        self.__ws_endpoint_on_open_vars_client['reqType'] = 'sub'
        self.__ws_endpoint_on_open_vars_client['dataType'] = None

        self.__ws_endpoint_on_close_vars_client['id'] = self.__ws_endpoint_on_open_vars_client['id']
        self.__ws_endpoint_on_close_vars_client['reqType'] = 'unsub'
        self.__ws_endpoint_on_close_vars_client['dataType'] = None

        self.__api_data_vars = None
        self.__api_data_vars = []

        __ws_args_client = None
        __ws_args_client = []

        __api_args_client = None
        __api_args_client = []

        for stream in self.__ws_streams:
            interval = 'none'

            if stream['endpoint'] == 'order_book':
                __ws_args_client.append(stream['symbol'].replace("/","-").upper()\
                                             + '@depth' + '100')

            elif stream['endpoint'] == 'kline':
                interval = stream['interval']
                __ws_args_client.append(stream['symbol'].replace("/","-").upper()\
                                             + '@kline_'\
                                             + self.__get_interval_from_unified_interval(interval))

            elif stream['endpoint'] == 'trades':
                __data_add = None
                __data_add = {}
                __data_add['endpoint'] = 'trades'
                __data_add['symbol'] = stream['symbol'].replace("/","-").upper()

                self.__api_data_vars.append(__data_add)

            elif stream['endpoint'] == 'ticker':
                __data_add = None
                __data_add = {}
                __data_add['endpoint'] = 'ticker'
                __data_add['symbol'] = stream['symbol'].replace("/","-").upper()

                self.__api_data_vars.append(__data_add)

            __stream_index = self.get_stream_index(stream['endpoint'],\
                                                     stream['symbol'],\
                                                     interval=interval)
            self.__ws_temp_data[__stream_index] = None

        self.__ws_endpoint_on_open_vars_client['dataType'] = __ws_args_client
        self.__ws_endpoint_on_close_vars_client['dataType'] = __ws_args_client

        result = {}
        result['ws_endpoint_url'] = self.__ws_endpoint_url
        result['ws_endpoint_on_open_vars'] = self.__ws_endpoint_on_open_vars
        result['ws_endpoint_on_close_vars'] = self.__ws_endpoint_on_close_vars

        return result

    def __get_data_from_api(self):
        """
        __get_data_from_api
        =================================
            This function get data from Rest API and send data to a local websocket server
            Requests limit 500 per minute and 1500 per 5 minutes -> 5 per second
            In our case 2 requests per second max.

                :return bool: True if all OK
        """

        result = True
        __min_time = 0.5
        __timeout = 9

        __to_stop = False

        with self.__lock:
            __to_stop = self.__is_stopped

        while not __to_stop:
            if self.__api_data_vars is not None\
                and isinstance(self.__api_data_vars, list)\
                and len(self.__api_data_vars) > 0:
                for __api_data_var in self.__api_data_vars:
                    __time_sleep = 0
                    __time_ini = time.time_ns()
                    if __api_data_var is not None and isinstance(__api_data_var, dict)\
                        and 'endpoint' in __api_data_var and 'symbol' in __api_data_var:
                        __data_url = None
                        __data_type = __api_data_var['symbol'] + '@' + __api_data_var['endpoint']

                        if __api_data_var['endpoint'] == 'trades':

                            __l_endpoint = '/openApi/' + str(self.__trading_type).lower()\
                                + '/v1/market/' + str(__api_data_var['endpoint'])\
                                + '?symbol=' + str(__api_data_var['symbol']) + '&limit=100'
                            __data_url = self.__url_api + __l_endpoint

                        elif __api_data_var['endpoint'] == 'ticker':
                            __timestamp = str(round(time.time_ns() / 1000000))
                            __l_endpoint = '/openApi/' + str(self.__trading_type).lower()\
                                + '/v1/' + str(__api_data_var['endpoint']) + '/24hr?timestamp='\
                                + __timestamp + '&symbol=' + str(__api_data_var['symbol'])
                            __data_url = self.__url_api + __l_endpoint

                        if __data_url is not None:
                            __message = (
                                ccf.file_get_contents_url(__data_url, 'b', None, {}, __timeout)
                            )

                            if __message is not None and ccf.is_json(__message):
                                __message = json.loads(__message)
                                if __message is not None\
                                    and isinstance(__message, dict)\
                                    and 'code' in __message\
                                    and 'data' in __message:
                                    __message_out = None
                                    __message_out = {}
                                    __message_out['code'] = __message['code']
                                    __message_out['data'] = __message['data']
                                    __message_out['timestamp'] = __message['timestamp']
                                    __message_out['dataType'] = __data_type
                                    __message_out['success'] = True
                                    __message_out = json.dumps(__message_out)
                                    self.__ws_server.send_message_to_all(__message_out)

                    __time_end = time.time_ns()
                    __time_diff = (__time_end - __time_ini) / 1000000000
                    __time_sleep = round((__min_time - __time_diff),3)

                    if __time_sleep > 0:
                        time.sleep(__time_sleep)
            else:
                time.sleep(1)

            with self.__lock:
                __to_stop = self.__is_stopped

        return result

    def manage_websocket_message_order_book(self, data, symbol):
        """
        manage_websocket_message_order_book
        ===================================
            This function manage websocket message and normalize result
            data for order_book endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        symbol = self.get_unified_symbol_from_symbol(symbol)
        __stream_index = self.get_stream_index('order_book', symbol)

        __temp_data = data
        __proc_data = False

        if __temp_data is not None and isinstance(__temp_data,dict)\
            and 'dataType' in __temp_data and 'data' in __temp_data\
            and isinstance(__temp_data['data'],dict):
            __bids = __temp_data['data']['bids']
            __asks = __temp_data['data']['asks']
            __asks.reverse()

            __message_out = None
            __message_out = {}
            __message_out['endpoint'] = 'order_book'
            __message_out['exchange'] = self.__exchange
            __message_out['symbol'] = symbol
            __message_out['interval'] = None
            __message_out['last_update_id'] = time.time_ns()
            __message_out['diff_update_id'] = 0
            __message_out['bids'] = __bids[:self.__result_max_len]
            __message_out['asks'] = __asks[:self.__result_max_len]
            __message_out['type'] = 'snapshot'
            __current_datetime = datetime.datetime.utcnow()
            __current_timestamp = __current_datetime.strftime("%s.%f")
            __current_datetime = __current_datetime.strftime("%Y-%m-%d %H:%M:%S.%f")
            __message_out['timestamp'] = __current_timestamp
            __message_out['datetime'] = __current_datetime

            result = __message_out
            self.__ws_temp_data[__stream_index] = __temp_data

        return result

    def manage_websocket_message_kline(self, data, symbol, interval):
        """
        manage_websocket_message_kline
        ==============================
            This function manage websocket message and normalize result
            data for kline endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        symbol = self.get_unified_symbol_from_symbol(symbol)
        interval = self.get_unified_interval_from_interval(interval)

        __stream_index = self.get_stream_index('kline', symbol, interval)

        if self.__ws_temp_data[__stream_index] is None\
            or not isinstance(self.__ws_temp_data[__stream_index], dict):
            self.__ws_temp_data[__stream_index] = {}

        if __temp_data is not None and isinstance(__temp_data,dict)\
            and 'code' in __temp_data and int(__temp_data['code']) == 0\
            and 'dataType' in __temp_data:
            if 'data' in __temp_data and isinstance(__temp_data['data'],dict)\
                and 'E' in __temp_data['data'] and 'K' in __temp_data['data']\
                and isinstance(__temp_data['data']['K'],dict):

                __out_interval = '1m' # Only available for 1 minute

                __message_add = None
                __message_add = {}
                __message_add['endpoint'] = 'kline'
                __message_add['exchange'] = self.__exchange
                __message_add['symbol'] = symbol
                __message_add['interval'] = __out_interval
                __message_add['last_update_id'] = __temp_data['data']['E']
                __message_add['open_time'] = int(__temp_data['data']['K']['t'])
                __message_add['close_time'] = int(__temp_data['data']['K']['T'])
                __message_add['open_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",\
                    time.gmtime(int(round(__message_add['open_time']/1000))))
                )
                __message_add['close_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",\
                    time.gmtime(int(round(__message_add['close_time']/1000))))
                )
                __message_add['open'] = __temp_data['data']['K']['o']
                __message_add['close'] = __temp_data['data']['K']['c']
                __message_add['hight'] = __temp_data['data']['K']['h']
                __message_add['low'] = __temp_data['data']['K']['l']
                __message_add['volume'] = __temp_data['data']['K']['v']
                __message_add['is_closed'] = None

                self.__ws_temp_data[__stream_index][int(__message_add['open_time'])] = __message_add

                while len(self.__ws_temp_data[__stream_index]) > self.__data_max_len:
                    __first_key = min(list(self.__ws_temp_data[__stream_index].keys()))
                    __nc = self.__ws_temp_data[__stream_index].pop(__first_key,None)

                __message_out = list(self.__ws_temp_data[__stream_index].values())

                result = __message_out[:self.__result_max_len]

        return result

    def manage_websocket_message_trades(self, data, symbol):
        """
        manage_websocket_message_trades
        ===============================
            This function manage websocket message and normalize result
            data for trades endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        symbol = self.get_unified_symbol_from_symbol(symbol)
        __stream_index = self.get_stream_index('trades', symbol)

        if __temp_data is not None and isinstance(__temp_data, dict)\
            and 'code' in __temp_data and  int(__temp_data['code']) == 0:
            if 'timestamp' in __temp_data and 'data' in __temp_data\
                and isinstance(__temp_data['data'],list)\
                and len(__temp_data['data']) > 0:

                if self.__ws_temp_data[__stream_index] is None:
                    self.__ws_temp_data[__stream_index] = {}

                for i in range(len(__temp_data['data']) - 1,-1,-1):
                    __message_add = None
                    __message_add = {}
                    __message_add['endpoint'] = 'trades'
                    __message_add['exchange'] = self.__exchange
                    __message_add['symbol'] = symbol
                    __message_add['interval'] = None
                    __message_add['event_time'] = __temp_data['timestamp']
                    __message_add['trade_id'] = str(__temp_data['data'][i]['id'])
                    __message_add['price'] = str(__temp_data['data'][i]['price'])
                    __message_add['quantity'] = str(__temp_data['data'][i]['qty'])
                    __message_add['trade_time'] = int(__temp_data['data'][i]['time'])
                    __message_add['trade_time_date'] = (
                        time.strftime("%Y-%m-%d %H:%M:%S",\
                        time.gmtime(int(round(__message_add['trade_time']/1000))))\
                            + '.' + str(round(math.modf(\
                                round(__message_add['trade_time']/1000,3))[0]*1000)\
                                    ).rjust(3,'0')
                    )

                    __side_of_taker = 'BUY'
                    if __temp_data['data'][i]['buyerMaker']:
                        __side_of_taker = 'SELL'
                    __message_add['side_of_taker'] = __side_of_taker

                    self.__ws_temp_data[__stream_index][int(__message_add['trade_id'])] = (
                        __message_add
                    )

                    while len(self.__ws_temp_data[__stream_index]) > self.__data_max_len:
                        __first_key = min(list(self.__ws_temp_data[__stream_index].keys()))
                        __nc = self.__ws_temp_data[__stream_index].pop(__first_key,None)

                    __message_out = list(self.__ws_temp_data[__stream_index].values())
                    #__message_out.reverse()
                    __message_out = __message_out[:self.__result_max_len]

                    result = __message_out

        return result

    def manage_websocket_message_ticker(self, data, symbol):
        """
        manage_websocket_message_ticker
        ===============================
            This function manage websocket message and normalize result
            data for ticker endpoint.

                :param data: dict.
                :return dict: Return dict with normalized data.
        """
        result = None

        __temp_data = data
        __proc_data = False

        symbol = self.get_unified_symbol_from_symbol(symbol)
        __stream_index = self.get_stream_index('ticker', symbol)

        if __temp_data is not None and isinstance(__temp_data,dict)\
            and 'code' in __temp_data and  int(__temp_data['code']) == 0:
            if 'timestamp' in __temp_data and 'data' in __temp_data\
                and isinstance(__temp_data['data'],list)\
                and len(__temp_data['data']) > 0:
                self.__ws_temp_data[__stream_index] = __temp_data

                i = 0
                __message_add = None
                __message_add = {}
                __message_add['endpoint'] = 'ticker'
                __message_add['exchange'] = self.__exchange
                __message_add['symbol'] = symbol
                __message_add['interval'] = None
                __message_add['event_type'] = '24hrTicker'
                __message_add['event_time'] = __temp_data['timestamp']
                __message_add['event_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",\
                        time.gmtime(int(round(int(__message_add['event_time'])/1000))))\
                        + '.' + str(round(math.modf(round(__message_add['event_time']/\
                            1000,3))[0]*1000)).rjust(3,'0')
                )
                __message_add['price_change'] = None
                __message_add['price_change_percent'] = None
                __message_add['weighted_average_price'] = None
                __message_add['first_trade_before_the_24hr_rolling_window'] = None
                __message_add['last_price'] = str(__temp_data['data'][i]['lastPrice'])
                __message_add['last_quantity'] = None
                __message_add['best_bid_price'] = None
                __message_add['best_bid_quantity'] = None
                __message_add['best_ask_price'] = None
                __message_add['best_ask_quantity'] = None
                __message_add['open_price'] = str(__temp_data['data'][i]['openPrice'])
                __message_add['high_price'] = str(__temp_data['data'][i]['highPrice'])
                __message_add['low_price'] = str(__temp_data['data'][i]['lowPrice'])
                __message_add['total_traded_base_asset_volume'] = (
                    str(__temp_data['data'][i]['volume'])
                )
                __message_add['total_traded_quote_asset_volume'] = (
                    str(__temp_data['data'][i]['quoteVolume'])
                )
                __message_add['statistics_open_time'] = (
                    __temp_data['data'][i]['openTime']
                )
                __message_add['statistics_open_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",\
                        time.gmtime(int(round(__message_add['statistics_open_time']/\
                            1000)))) + '.' +\
                                str(round(math.modf(\
                                    round(__message_add['statistics_open_time']/\
                                        1000,3))[0]*1000)).rjust(3,'0')
                )
                __message_add['statistics_close_time'] = (
                    __temp_data['data'][i]['closeTime']
                )
                __message_add['statistics_close_time_date'] = (
                    time.strftime("%Y-%m-%d %H:%M:%S",time.gmtime(int(\
                        round(__message_add['statistics_close_time']/1000))))\
                        + '.' + str(round(math.modf(\
                            round(__message_add['statistics_close_time']/\
                            1000,3))[0]*1000)).rjust(3,'0')
                )
                __message_add['total_number_of_trades'] = None

                __message_out = __message_add

                result = __message_out

        return result


    def manage_websocket_message(self, ws, message_in):
        """
        manage_websocket_message
        ========================
            This function manage websocket message and normalize result data.

                :param ws: WebSocketApp instance.
                :param message_in: str message from websocket.

                :return dict: Return dict with normalized data.
        """

        result = None

        try:
            if isinstance(message_in,bytes):
                message_in = gzip.GzipFile(fileobj=io.BytesIO(message_in), mode='rb')
                message_in = message_in.read()
                message_in = message_in.decode('utf-8')

            if message_in == 'Ping':
                ws.send('Pong')
                result = message_in
            else:
                if ccf.is_json(message_in):
                    __temp_data = json.loads(message_in)
                    __endpoint = 'NONE'
                    __symbol = None
                    __interval = None
                    __message_out = None

                    if __temp_data is not None\
                        and isinstance(__temp_data, dict)\
                        and 'code' in __temp_data\
                        and 'data' in __temp_data\
                        and 'dataType' in __temp_data\
                        and 'success' in __temp_data\
                        and __temp_data['success']:

                        __tmp_endpoint = 'NONE'

                        if '@' in __temp_data['dataType']:
                            __input_tmp = __temp_data['dataType'].split('@')
                            if __input_tmp is not None\
                                and isinstance(__input_tmp, list)\
                                and len(__input_tmp) >=2:
                                __symbol = __input_tmp[0]
                                __tmp_endpoint = __input_tmp[1]

                        if __tmp_endpoint.startswith('depth'):
                            __endpoint = 'order_book'
                        elif __tmp_endpoint.startswith('kline') and '_' in __tmp_endpoint:
                            __tmp_interval = __tmp_endpoint.split('_')
                            if __tmp_interval is not None\
                                and isinstance(__tmp_interval, list)\
                                and len(__tmp_interval) >=2:
                                __endpoint = 'kline'
                                __interval = __tmp_interval[1]
                        elif __tmp_endpoint.startswith('trades'):
                            __endpoint = 'trades'
                        elif __tmp_endpoint.startswith('ticker'):
                            __endpoint = 'ticker'

                        if __endpoint == 'order_book':
                            __message_out = (
                                self.manage_websocket_message_order_book(__temp_data, __symbol)
                            )

                            if __message_out is not None:

                                result = {}
                                result['data'] = __message_out
                                result['min_proc_time_ms'] = 0
                                result['max_proc_time_ms'] = 0

                        elif __endpoint == 'kline':
                            __message_out = (
                                self.manage_websocket_message_kline(__temp_data,\
                                                                    __symbol,\
                                                                    __interval)
                            )

                            if __message_out is not None:
                                result = {}
                                result['data'] = __message_out
                                result['min_proc_time_ms'] = 0
                                result['max_proc_time_ms'] = 0

                        elif __endpoint == 'trades':
                            __message_out = (
                                self.manage_websocket_message_trades(__temp_data, __symbol)
                            )

                            if __message_out is not None:
                                result = {}
                                result['data'] = __message_out
                                result['min_proc_time_ms'] = 0
                                result['max_proc_time_ms'] = 0

                        elif __endpoint == 'ticker':
                            __message_out = (
                                self.manage_websocket_message_ticker(__temp_data, __symbol)
                            )

                            if __message_out is not None:
                                result = {}
                                result['data'] = __message_out
                                result['min_proc_time_ms'] = 0
                                result['max_proc_time_ms'] = 0

        except Exception as exc: # pylint: disable=broad-except
            print('EXCEPTION: ' + str(exc))

        return result



"""
Ccxw - CryptoCurrency eXchange Websocket Library
Thread safe dict

Author: Ricardo Marcelo Alvarez
Date: 2023-10-31
"""

from threading import Lock

class DictSafeThread():
    """
    Thread safe dict
    ===================================================================
        This class contains helper functions for the Ccxw class.
    """

    def __init__(self, dict_in=None):
        self.__lock = Lock()
        self._data = None

        with self.__lock:
            if dict_in is None:
                self._data = {}
            elif isinstance(dict_in, dict):
                self._data = dict_in
            else:
                raise ValueError('Input data is not None or dict')

    def __del__(self):
        pass

    def __enter__(self):
        self.__lock.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.__lock.__exit__(exc_type, exc_value, traceback)

    def __str__(self):
        result = None
        with self.__lock:
            result = str(self._data)

        return result

    def __repr__(self):

        result = None
        with self.__lock:
            result = repr(self._data)

        return result

    def __getitem__(self, key):
        result = None

        with self.__lock:
            if key in self._data:
                result = self._data[key]

        return result

    def __setitem__(self, key, value):
        result = None

        with self.__lock:
            self._data[key] = value

        return result

    def __delitem__(self, key):
        result = None

        with self.__lock:
            del self._data[key]

        return result

    def __len__(self):
        result = 0

        with self.__lock:
            result = len(self._data)

        return result

    def __iter__(self):
        result = None

        with self.__lock:
            result = self._data.keys()
            result = iter(result)

        return result

    def __contains__(self, key):
        result = False

        with self.__lock:
            result = key in self._data

        return result

    def __eq__(self, other):
        result = False

        with self.__lock:
            if isinstance(other, DictSafeThread):
                result = self._data == other._data

        return result

    def __ne__(self, other):
        result = False
        result = not self.__eq__(other)
        return result

    def __hash__(self):
        result = ''

        with self.__lock:
            result = hash(self.__lock)

        return hash(result)
